from typing import List, Literal
import traceback
import ujson

from fastapi import (APIRouter, Depends, Request, Form, status,
                    Response, UploadFile, File, HTTPException)

from promptengineers.core.exceptions import ValidationException
from promptengineers.core.interfaces.controllers import IController
from promptengineers.models.request import RequestDataLoader, RequestMultiLoader
from promptengineers.models.response import (ResponseFileLoader, ResponseCreateVectorStore,
									ResponseListPineconeVectorStores)
from promptengineers.fastapi.controllers import VectorSearchController, AuthController
from promptengineers.retrieval.factories import RetrievalFactory, EmbeddingFactory, LoaderFactory
from promptengineers.retrieval.strategies import VectorstoreContext
from promptengineers.core.utils import logger
from promptengineers.core.exceptions import NotFoundException

TAG = "Retrieval"
router = APIRouter()
auth_controller = AuthController()

def get_controller(request: Request) -> IController:
	try:
		return VectorSearchController(request=request, user_repo=request.state.user_repo)
	except NotFoundException as e:
		# Handle specific NotFoundException with a custom message or logging
		logger.warn(f"Failed to initialize HistoryController: {str(e)}")
		raise HTTPException(status_code=404, detail=f"Initialization failed: {str(e)}") from e
	except Exception as e:
		# Catch all other exceptions
		logger.error(f"Unexpected error initializing HistoryController: {str(e)}")
		raise HTTPException(status_code=500, detail="Internal server error") from e

#################################################
# Create vectorstore from files
#################################################
@router.post(
	"/vectorstores",
	name='retrieval_vectorstore_create',
	response_model=ResponseCreateVectorStore,
	tags=[TAG]
)
async def create_vectorstore(
	body: RequestDataLoader,
	controller: VectorSearchController = Depends(get_controller),
):
	"""File Loader endpoint."""
	logger.debug('[POST /vectorstores] Body: %s', str(body))
	try:
		tokens = await controller.user_repo.find_token(
			controller.user_id, 
			['OPENAI_API_KEY', 'PINECONE_API_KEY', 'PINECONE_ENV', 'PINECONE_INDEX', 'REDIS_URL']
		)

		# Generate Embeddings
		embedding = EmbeddingFactory(body.embedding, tokens.get('OPENAI_API_KEY'))

		# Generate Provider Keys
		if body.provider == 'redis':
			provider_keys={
				'redis_url': tokens.get('REDIS_URL'),
				'index_name': f"{controller.request.state.user_id}::{body.index_name}",
			}
		elif body.provider == 'pinecone':
			provider_keys = {
				'api_key': tokens.get('PINECONE_API_KEY'),
				'env': tokens.get('PINECONE_ENV'),
				'index_name': tokens.get('PINECONE_INDEX'),
				'namespace': f"{controller.request.state.user_id}::{body.index_name}",
			}
		else:
			raise HTTPException(
				status_code=400,
				detail=f"Invalid retrieval provider: {body.provider}"
			)
		
		# Generate Loaders
		all_loaders = []
		for loader_config in body.loaders:
			try:
				loader = LoaderFactory.create(loader_type=loader_config['type'], loader_config=loader_config)
				all_loaders.append(loader)
			except ValueError as e:
				raise HTTPException(
					status_code=500,
					detail=f"Error creating loader: {e}"
				)
			except Exception as e:
				raise HTTPException(
					status_code=500,
					detail=f"Unexpected error: {e}"
				)
			
		# Generate Retrieval Provider
		try:
			retrieval_provider = RetrievalFactory(
				provider=body.provider,
				embeddings=embedding.create_embedding(),
				provider_keys=provider_keys
			)
			# Create a vector store service context
			vectostore_service = VectorstoreContext(retrieval_provider.create_strategy())
			vectostore_service.add(all_loaders)
		except Exception as e:
			raise HTTPException(
				status_code=500,
				detail=f"Error processing document: {e}"
			)

		## Format Response
		data = ujson.dumps({
			"message": 'Vectorstore Created!',
			'vectorstore': body.index_name,
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=201
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		tb = traceback.format_exc()
		logger.error("[routes.vectorstores.create_vectorstore] HTTPException: %s\n%s", err, tb)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.vectorstores.create_vectorstore] Exception: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Create vectorstore from files
#################################################
@router.post(
	"/vectorstores/file",
	name='retrieval_vectorstore_file_create',
	response_model=ResponseFileLoader,
	tags=[TAG]
)
async def create_vectorstore_from_file(
	index_name: str = Form(...),
	provider: Literal['pinecone', 'redis'] = Form(...),
	embedding: Literal['text-embedding-ada-002', 'llama2:7b', 'llama2'] = Form(...),
	files: List[UploadFile] = File(...),
	controller: VectorSearchController = Depends(get_controller),
):
	"""File Loader endpoint."""
	try:
		await controller.create_vectorstore_from_files(
			provider,
			index_name,
			embedding,
			files
		)

		## Format Response
		data = ujson.dumps({
			"message": 'Vectorstore Created!',
			'vectorstore': index_name,
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=201
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("HTTPException: %s", err.detail)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.vectorstores.create_vectorstore_from_file]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err


#################################################
# Create vectorstore
#################################################
@router.post(
	"/vectorstores/multi",
	name='retrieval_vectorstore_multi_create',
	response_model=ResponseCreateVectorStore,
	tags=[TAG],
	# include_in_schema=False # TODO: Needs some work
)
async def create_vectorstore_from_multiple_sources(
	request: Request,
	body: RequestMultiLoader,
	controller: VectorSearchController = Depends(get_controller),
):
	# Get Body
	body = await request.json()
	logger.debug('[POST /vectorstores/multi] Body: %s', str(body))

	try:
		await controller.create_multi_loader_vectorstore(body)
		## Format Response
		data = ujson.dumps({
			'message': 'Vectorstore Created!',
			'data': body
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=201
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("%s", err.detail, stack_info=True)
		raise
	except Exception as err:
		logger.error("%s", err, stack_info=True)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err


######################################
# Retrieve Vector Stores
######################################
@router.get(
	"/vectorstores/pinecone",
	name='retrieval_vectorstore_pinecone_list',
	response_model=ResponseListPineconeVectorStores,
	tags=[TAG]
)
async def list_pinecone_vectorstores(controller: VectorSearchController = Depends(get_controller)):
	try:
		result = await controller.retrieve_pinecone_vectorstores()
		# Format Response
		data = ujson.dumps({
			**result
		})
		logger.debug("List Pinecone Vectorstores: %s", data)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("HTTPException: %s", err.detail)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.retrieval.list_pinecone_vectorstores]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

######################################
# Delete Vector Store
######################################
@router.delete(
	"/vectorstores/pinecone",
	name='retrieval_vectorstore_pinecone_delete',
	status_code=status.HTTP_204_NO_CONTENT,
	tags=[TAG]
)
async def delete_pinecone_vectorstore(
	prefix: str or None = None,
	controller: VectorSearchController = Depends(get_controller)
):
	try:
		await controller.delete_pinecone_vectorstore(prefix)
		return Response(status_code=204)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except Exception as err:
		raise HTTPException(status_code=404, detail=str(err)) from err

######################################
# Retrieve Vector Stores
######################################
@router.get(
	"/vectorstores/redis",
	name='retrieval_vectorstore_redis_list',
	response_model=ResponseListPineconeVectorStores,
	tags=[TAG]
)
async def list_redis_vectorstores(controller: VectorSearchController = Depends(get_controller)):
	try:
		result = controller.retrieve_redis_vectorstores()
		# Format Response
		data = ujson.dumps({
			**result
		})
		logger.debug("List Redis Vectorstores: %s", data)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("HTTPException: %s", err.detail)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.vectorstores.list_redis_vectorstores]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

######################################
# Delete Vector Store
######################################
@router.delete(
	"/vectorstores/redis",
	name='retrieval_vectorstore_redis_delete',
	status_code=status.HTTP_204_NO_CONTENT,
	tags=[TAG]
)
async def delete_redis_vectorstore(
	prefix: str or None = None,
	controller: VectorSearchController = Depends(get_controller)
):
	try:
		controller.delete_redis_vectorstore(prefix)
		return Response(status_code=204)
	except ValidationException as err:
		logger.warning("ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except Exception as err:
		raise HTTPException(status_code=404, detail=str(err)) from err


from typing import Optional
from pydantic import BaseModel

class Retrieval(BaseModel):
    """Contains the information needed to document retrieval augmented generation."""

    provider: Optional[str] = None
    index_name: Optional[str] = None

    __config__ = {
		"json_schema_extra": {
            "example": {
                "provider": "pinecone",
                "index_name": "Formio",
            }
        }
    }


from promptengineers.models.history import HistoryDocument, HISTORY_DOCUMENT

class ResponseHistoryIndex(HistoryDocument): # pylint: disable=too-few-public-methods
	histories: list[HistoryDocument]

	__config__ = {
		"json_schema_extra": {
			"example": {
				"histories": [HISTORY_DOCUMENT]
            }
		}
	}

class ResponseHistoryShow(HistoryDocument): # pylint: disable=too-few-public-methods
	history: HistoryDocument

	__config__ = {
		"json_schema_extra": {
			"example": {
				"history": HISTORY_DOCUMENT
            }
		}
	}


from typing import Dict

from promptengineers.retrieval.strategies import PineconeStrategy, RedisStrategy, VectorStoreStrategy

class RetrievalFactory:
    def __init__(
        self,
        provider: str,
        embeddings,
        provider_keys: Dict[str, str]
    ):
        """
        A factory for creating vector store strategies based on the given provider.

        :param provider: The name of the provider ('pinecone' or 'redis').
        :param embeddings: The embeddings to be used in the strategy.
        :param provider_keys: A dictionary containing the keys for the provider configuration.
        """
        self.provider = provider
        self.embeddings = embeddings
        self.provider_keys = provider_keys
        self._validate_provider_keys()

    def create_strategy(self) -> VectorStoreStrategy:
        """
        Creates and returns the appropriate strategy object based on the provider.

        :return: An instance of PineconeStrategy or RedisStrategy.
        :raises ValueError: If the provider is not supported.
        """
        if self.provider == 'pinecone':
            return self._create_pinecone_strategy()
        elif self.provider == 'redis':
            return self._create_redis_strategy()
        else:
            raise ValueError(f"Invalid index provider: {self.provider}")

    def _create_pinecone_strategy(self) -> PineconeStrategy:
        """ Creates a PineconeStrategy with the appropriate configuration. """
        return PineconeStrategy(
            embeddings=self.embeddings,
            api_key=self.provider_keys['api_key'],
            env=self.provider_keys['env'],
            index_name=self.provider_keys['index_name'],
            namespace=self.provider_keys['namespace'],
        )

    def _create_redis_strategy(self) -> RedisStrategy:
        """ Creates a RedisStrategy with the appropriate configuration. """
        return RedisStrategy(
            redis_url=self.provider_keys['redis_url'],
            index_name=self.provider_keys['index_name'],
            embeddings=self.embeddings,
        )

    def _validate_provider_keys(self):
        """
        Validates that all required keys are present for the configured provider.
        :raises KeyError: If any required key is missing.
        """
        required_keys = {
            'pinecone': ['api_key', 'env', 'index_name', 'namespace'],
            'redis': ['redis_url', 'index_name']
        }

        if self.provider not in required_keys:
            raise ValueError(f"Unsupported provider: {self.provider}")

        missing_keys = [
            key for key in required_keys[self.provider] 
            if key not in self.provider_keys
        ]

        if missing_keys:
            raise KeyError(
                f"Missing keys for provider '{self.provider}': {', '.join(missing_keys)}"
            )


import ujson
import traceback

from fastapi import APIRouter, Depends, Request, Response, HTTPException
from fastapi.responses import StreamingResponse

from promptengineers.chat import (langchain_http_agent_chat, langchain_stream_agent_chat,
									langchain_http_retrieval_chat, langchain_stream_retrieval_chat)
from promptengineers.fastapi.controllers import ChatController
from promptengineers.core.exceptions import ValidationException
from promptengineers.retrieval.factories import RetrievalFactory, EmbeddingFactory
from promptengineers.models.request import ReqBodyChat, ReqBodyAgentChat, ReqBodyVectorstoreChat
from promptengineers.models.response import (ResponseChat, ResponseAgentChat, ResponseVectorstoreChat,
									RESPONSE_STREAM_AGENT_CHAT, RESPONSE_STREAM_VECTORSTORE_CHAT,
									RESPONSE_STREAM_CHAT)
from promptengineers.retrieval.strategies import VectorstoreContext
from promptengineers.core.utils import logger
from promptengineers.llms.utils import gather_tools, retrieve_system_message
from promptengineers.tools.utils import format_agent_actions
from promptengineers.core.exceptions import NotFoundException

router = APIRouter()
TAG = "Chat"

def get_controller(request: Request) -> ChatController:
	try:
		return ChatController(
			request=request, 
			user_repo=request.state.user_repo, 
			available_tools=request.state.available_tools
		)
	except NotFoundException as e:
		# Handle specific NotFoundException with a custom message or logging
		logger.warn(f"Failed to initialize HistoryController: {str(e)}")
		raise HTTPException(status_code=404, detail=f"Initialization failed: {str(e)}") from e
	except Exception as e:
		# Catch all other exceptions
		logger.error(f"Unexpected error initializing HistoryController: {str(e)}")
		raise HTTPException(status_code=500, detail="Internal server error") from e

#################################################
# ChatGPT
#################################################
@router.post(
	"/chat",
	tags=[TAG],
	name='chat',
	response_model=ResponseChat,
	responses={
		200: {
			"content": {
				"text/event-stream": {
					"example": RESPONSE_STREAM_CHAT
				}
			}
		}
	},
)
async def chat(
	body: ReqBodyChat,
	chat_controller: ChatController = Depends(get_controller),
):
	"""Chat endpoint."""
	try:
		logger.debug('[POST /chat] Query: %s', str(body))
		# You can use the stream variable in your function as needed
		if not body.stream:
			# Format Response
			result, cb = await chat_controller.langchain_http_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature
			)
			data = ujson.dumps({
				'message': result,
				'usage': {
					'total_tokens': cb.total_tokens,
					'prompt_tokens': cb.prompt_tokens,
					'completion_tokens': cb.completion_tokens,
					'total_cost': cb.total_cost,
					'successful_requests': cb.successful_requests
				},
			})
			logger.debug('[POST /chat] Result: %s', str(data))
			return Response(
				content=data,
				media_type='application/json',
				status_code=200
			)

		return StreamingResponse(
			chat_controller.langchain_stream_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature
			),
			headers={
				"Cache-Control": "no-cache",
				"Connection": "keep-alive",
				"Content-Type": "text/event-stream",
			}
		)
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.chat.chat] Exception: %s\n%s", err, tb)
		raise HTTPException(status_code=500, detail="Internal Server Error") from err



#################################################
# Langchain Agent
#################################################
@router.post(
	"/chat/agent",
	tags=[TAG],
	name='chat_agent',
	response_model=ResponseAgentChat,
	responses={
		200: {
			"content": {
				"text/event-stream": {
					"example": RESPONSE_STREAM_AGENT_CHAT
				}
			}
		}
	},
)
async def agent(
	body: ReqBodyAgentChat,
	chat_controller: ChatController = Depends(get_controller),
):
	"""Chat endpoint."""
	try:
		tokens = await chat_controller.user_repo.find_token(
			chat_controller.user_id, 
			['OPENAI_API_KEY', 'PINECONE_API_KEY', 'PINECONE_ENV', 'PINECONE_INDEX', 'REDIS_URL']
		)

		vectorstore = None
		if body.retrieval.provider and body.retrieval.index_name:

			# Generate Embeddings
			embedding = EmbeddingFactory(body.model, tokens.get('OPENAI_API_KEY'))

			if body.retrieval.provider == 'redis':
				provider_keys={
					'redis_url': tokens.get('REDIS_URL'),
					'index_name': f"{chat_controller.request.state.user_id}::{body.retrieval.index_name}",
				}
			elif body.retrieval.provider == 'pinecone':
				provider_keys = {
					'api_key': tokens.get('PINECONE_API_KEY'),
					'env': tokens.get('PINECONE_ENV'),
					'index_name': tokens.get('PINECONE_INDEX'),
					'namespace': f"{chat_controller.request.state.user_id}::{body.retrieval.index_name}",
				}
			else:
				raise HTTPException(
					status_code=400,
					detail=f"Invalid retrieval provider: {body.retrieval.provider}"
				)

			retrieval_provider = RetrievalFactory(
				provider=body.retrieval.provider,
				embeddings=embedding.create_embedding(),
				provider_keys=provider_keys
			)
			vectostore_service = VectorstoreContext(retrieval_provider.create_strategy())
			vectorstore = vectostore_service.load()
		
		tools = gather_tools(
			tools=body.tools,
			available_tools=chat_controller.available_tools,
			vectorstore=vectorstore,
			plugins=body.plugins
		)
		if not tools:
			raise HTTPException(
				status_code=400,
				detail="No tools selected"
			)

		# Retrieve the system message
		system_message = retrieve_system_message(body.messages)
		# Attach the application user id to the system message
		system_message = system_message + '\n' + "USER_ID=" + str(chat_controller.user_id)
		body.messages[0]['content'] = system_message

		# You can use the stream variable in your function as needed
		if not body.stream:
			# Format Response
			result, cb = await langchain_http_agent_chat(
				messages=body.messages,
				model=body.model,
				tools=tools,
				temperature=body.temperature,
				openai_api_key=tokens.get('OPENAI_API_KEY'),
			)
			data = ujson.dumps({
				'message': result['output'],
				'actions': format_agent_actions(result['intermediate_steps']),
				'usage': {
					'total_tokens': cb.total_tokens,
					'prompt_tokens': cb.prompt_tokens,
					'completion_tokens': cb.completion_tokens,
					'total_cost': cb.total_cost,
					'successful_requests': cb.successful_requests
				},
			})
			logger.debug('[POST /chat/agent] Result: %s', str(data))
			return Response(
				content=data,
				media_type='application/json',
				status_code=200
			)

		return StreamingResponse(
			langchain_stream_agent_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature,
				tools=tools,
				openai_api_key=tokens.get('OPENAI_API_KEY'),
			),
			headers={
				"Cache-Control": "no-cache",
				"Connection": "keep-alive",
				"Content-Type": "text/event-stream",
			}
		)
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.chat.vector_search] Exception: %s\n%s", err, tb)
		raise HTTPException(status_code=500, detail="Internal Server Error") from err

#################################################
# Langchain Vectorstore Route
#################################################
@router.post(
	"/chat/vectorstore",
	tags=[TAG],
	name='chat_vectorstore',
	response_model=ResponseVectorstoreChat,
	responses={
		200: {
			"content": {
				"text/event-stream": {
					"example": RESPONSE_STREAM_VECTORSTORE_CHAT
				}
			}
		}
	},
)
async def vector_search(
	body: ReqBodyVectorstoreChat,
	chat_controller: ChatController = Depends(get_controller),
):
	"""Chat Vectorstore endpoint."""
	try:
		# Log Context Details
		logger.debug('[POST /chat/vectorstore] Query: %s', str(body))

		# Retrieve User Tokens
		tokens = await chat_controller.user_repo.find_token(
			chat_controller.user_id, 
			['OPENAI_API_KEY', 'PINECONE_API_KEY', 'PINECONE_ENV', 'PINECONE_INDEX', 'REDIS_URL']
		)

		# Generate Embeddings
		embedding = EmbeddingFactory(body.model, tokens.get('OPENAI_API_KEY'))

		if body.provider == 'redis':
			provider_keys={
				'redis_url': tokens.get('REDIS_URL'),
				'index_name': f"{chat_controller.request.state.user_id}::{body.vectorstore}",
			}
		elif body.provider == 'pinecone':
			provider_keys = {
				'api_key': tokens.get('PINECONE_API_KEY'),
				'env': tokens.get('PINECONE_ENV'),
				'index_name': tokens.get('PINECONE_INDEX'),
				'namespace': f"{chat_controller.request.state.user_id}::{body.vectorstore}",
			}
		else:
			raise HTTPException(
				status_code=400,
				detail=f"Invalid retrieval provider: {body.provider}"
			)

		retrieval_provider = RetrievalFactory(
			provider=body.provider,
			embeddings=embedding.create_embedding(),
			provider_keys=provider_keys
		)
		vectostore_service = VectorstoreContext(retrieval_provider.create_strategy())
		vectorstore = vectostore_service.load()

		# Check if the retrieved file is empty
		if not vectorstore:
			raise HTTPException(
				status_code=404,
				detail=f"Vectorstore {body.vectorstore} not found"
			)

		# You can use the stream variable in your function as needed
		if not body.stream:
			# Format Response
			result, cb = await langchain_http_retrieval_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature,
				vectorstore=vectorstore,
				openai_api_key=tokens.get('OPENAI_API_KEY'),
			)
			formatted_docs = []
			for doc in result['source_documents']:
				formatted_docs.append({
					'page_content': doc.page_content,
					'metadata': doc.metadata,
				})
			data = ujson.dumps({
				'message': result['answer'],
				'documents': formatted_docs,
				'usage': {
					'total_tokens': cb.total_tokens,
					'prompt_tokens': cb.prompt_tokens,
					'completion_tokens': cb.completion_tokens,
					'total_cost': cb.total_cost,
					'successful_requests': cb.successful_requests
				},
			})
			logger.debug('[POST /chat/vectorstore] Result: %s', str(data))
			return Response(
				content=data,
				media_type='application/json',
				status_code=200
			)

		# Process Query
		return StreamingResponse(
			langchain_stream_retrieval_chat(
				messages=body.messages,
				model=body.model,
				temperature=body.temperature,
				vectorstore=vectorstore,
				openai_api_key=tokens.get('OPENAI_API_KEY'),
			),
			media_type="text/event-stream"
		)
	except ValidationException as err:
		logger.warning("[routes.chat.vector_search] ValidationException: %s", err)
		raise HTTPException(
			status_code=400,
			detail=str(err)
		) from err
	except HTTPException as err:
		logger.error("[routes.chat.vector_search] HTTPException: %s", err.detail)
		raise
	except Exception as err:
		tb = traceback.format_exc()
		logger.error("[routes.chat.vector_search] Exception: %s\n%s", err, tb)
		raise HTTPException(status_code=500, detail="Internal Server Error") from err



from promptengineers.models.history import History, HISTORY

class ReqBodyHistory(History): # pylint: disable=too-few-public-methods

	__config__ = {
		"json_schema_extra": {
			"example": HISTORY
		}
	}


from typing import Optional
from pydantic import BaseModel


class PromptSystem(BaseModel): # pylint: disable=too-few-public-methods
	"""Prompt System Template"""
	title: str
	system: str
	variables: Optional[dict] = None

	__config__ = {
		"json_schema_extra": {
			"example": {
				"title": "Helpful Assistant Prompt",
				"system": "You are a helpful {ASSISTANT_TYPE} assistant. You have have been given context to {INDEX_NAME} to answer user questions.",
				"variables": {
					"ASSISTANT_TYPE": 'document retreival',
					"INDEX_NAME": 'Stripe API docs'
                }
			}
		}
	}


"""Chat Functions"""
from .default import langchain_http_chat, langchain_stream_chat
from .retrieval import langchain_http_retrieval_chat, langchain_stream_retrieval_chat
from .agent import langchain_stream_agent_chat, langchain_http_agent_chat

__all__ = [
    'langchain_http_chat',
    'langchain_stream_chat',
    'langchain_http_retrieval_chat',
    'langchain_stream_retrieval_chat',
    'langchain_http_agent_chat',
    'langchain_stream_agent_chat'
]



from typing import Any
from promptengineers.core.config.llm import ACCEPTED_OLLAMA_MODELS, ACCEPTED_OPENAI_MODELS
from langchain_openai import OpenAIEmbeddings
from langchain.embeddings import OllamaEmbeddings

class EmbeddingFactory:
    def __init__(self, llm: str, token: str = None, base_url: str = None):
        """
        Initializes the EmbeddingFactory with specified parameters.

        :param llm: The language model to use for embeddings.
        :param token: The API token for accessing the language model service.
        :param base_url: The base URL for the language model service.
        :raises ValueError: If the language model is not supported.
        """
        if llm not in ACCEPTED_OPENAI_MODELS and llm not in ACCEPTED_OLLAMA_MODELS:
            raise ValueError(f"Invalid embedding model {llm}")

        self.llm = llm
        self.token = token
        self.base_url = base_url

    def create_embedding(self) -> Any:
        """
        Creates and returns an embeddings instance based on the stored language model.

        :return: An instance of OpenAIEmbeddings or OllamaEmbeddings.
        """
        return self._create_embedding(self.llm, self.token, self.base_url)

    @staticmethod
    def _create_embedding(llm: str, token: str, base_url: str) -> Any:
        """
        Creates and returns an embeddings instance based on the specified language model.

        :param llm: The language model to use for embeddings.
        :param token: The API token for accessing the language model service.
        :param base_url: The base URL for the language model service.
        :return: An instance of OpenAIEmbeddings or OllamaEmbeddings.
        :raises ValueError: If the language model is not supported.
        """
        if llm in ACCEPTED_OPENAI_MODELS:
            return OpenAIEmbeddings(openai_api_key=token, disallowed_special=())
        elif llm in ACCEPTED_OLLAMA_MODELS:
            return OllamaEmbeddings(base_url=base_url or 'http://127.0.0.1:11434')
        else:
            raise ValueError(f"Invalid embedding model {llm}")

    @classmethod
    def create_for_model(cls, llm: str, token: str = None, base_url: str = None) -> Any:
        """
        Creates and returns an embeddings instance for a specified language model.

        :param llm: The language model to use for embeddings.
        :param token: The API token for accessing the language model service.
        :param base_url: The base URL for the language model service.
        :return: An instance of OpenAIEmbeddings or OllamaEmbeddings.
        :raises ValueError: If the language model is not supported.
        """
        return cls._create_embedding(llm, token, base_url)
					 


"""Routes to get chat history"""
import json
import traceback

from fastapi import APIRouter, HTTPException, Response, Depends, Request, status
from promptengineers.core.exceptions import NotFoundException
from promptengineers.core.utils import logger
from promptengineers.fastapi.controllers import HistoryController
from promptengineers.models.request import ReqBodyHistory
from promptengineers.models.response import (ResponseHistoryShow, ResponseHistoryIndex,
                                            ResponseCreate, ResponseUpdate)
from promptengineers.mongo.utils import JSONEncoder

router = APIRouter()
TAG = "Chat"

def get_controller(request: Request) -> HistoryController:
    try:
        return HistoryController(request=request, user_repo=request.state.user_repo)
    except NotFoundException as e:
        # Handle specific NotFoundException with a custom message or logging
        logger.warn(f"Failed to initialize HistoryController: {str(e)}")
        raise HTTPException(status_code=404, detail=f"Initialization failed: {str(e)}") from e
    except Exception as e:
        # Catch all other exceptions
        logger.error(f"Unexpected error initializing HistoryController: {str(e)}")
        raise HTTPException(status_code=500, detail="Internal server error") from e

#################################################
# List Chat Histories
#################################################
@router.get(
	"/chat/history",
	tags=[TAG],
	name='history_list',
	response_model=ResponseHistoryIndex
)
async def list_chat_histories(
    page: int = 1,
    limit: int = 50,
    setting: str = None,
    controller: HistoryController = Depends(get_controller),
):
    """List histories"""
    try:
        # Forward the setting_id and any additional parameters to the controller's index method
        result = await controller.index(page, limit, setting=setting)
        # Format Response
        data = json.dumps({
            'histories': result
        }, cls=JSONEncoder)
        return Response(
            content=data,
            media_type='application/json',
            status_code=200
        )
    except HTTPException as err:
        logger.error(err.detail)
        raise
    except BaseException as err:
        tb = traceback.format_exc()
        logger.error("[list_chat_histories]: %s\n%s", err, tb)
        raise HTTPException(
            status_code=500,
            detail=f"An unexpected error occurred. {str(err)}"
        ) from err

#################################################
# Create Chat History
#################################################
@router.post(
	"/chat/history",
	tags=[TAG],
	name='history_create',
	response_model=ResponseCreate
)
async def create_chat_history(
	body: ReqBodyHistory,
	controller: HistoryController = Depends(get_controller)
):
	"""Create history"""
	try:
		result = await controller.create(body)
		# Format Response
		data = json.dumps({
			**result
		})
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error(err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[create_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Show Chat History
#################################################
@router.get(
	"/chat/history/{history_id}",
	tags=[TAG],
	name='history_show',
	response_model=ResponseHistoryShow,
)
async def show_chat_history(
    history_id: str,
    controller: HistoryController = Depends(get_controller),
):
	"""Show history"""
	try:
		result = await controller.show(history_id)

		# Format Response
		data = json.dumps({
			**result
		}, cls=JSONEncoder)
		return Response(
			content=data,
			media_type='application/json',
			status_code=200
		)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[show_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Update Chat History
#################################################
@router.put(
	"/chat/history/{history_id}",
	tags=[TAG],
	name='history_update',
	response_model=ResponseUpdate,
)
async def update_chat_history(
	history_id: str,
	body: ReqBodyHistory,
	controller: HistoryController = Depends(get_controller),
):
	"""Update history"""
	try:
		await controller.update(history_id, body)
		data = json.dumps({
			'message': f'History [{history_id}] updated successfully.'
		})
		# Format Response
		return Response(status_code=200, content=data)
	except HTTPException as err:
		logger.error("%s", err.detail)
		raise
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[update_chat_history]: %s\n%s", err, tb)
		raise HTTPException(
			status_code=500,
			detail=f"An unexpected error occurred. {str(err)}"
		) from err

#################################################
# Delete Chat History
#################################################
@router.delete(
	"/chat/history/{history_id}",
	tags=[TAG],
	name='history_delete',
	status_code=status.HTTP_204_NO_CONTENT,
)
async def delete_chat_history(
	history_id: str,
	controller: HistoryController = Depends(get_controller),
):
	"""Deletes history"""
	try:
		await controller.delete(history_id)
		# Format Response
		return Response(status_code=204)
	except BaseException as err:
		tb = traceback.format_exc()
		logger.error("[delete_chat_history]: %s\n%s", err, tb)
		raise HTTPException(status_code=404, detail=str(err)) from err


"""Retrieval Chat"""
from typing import AsyncIterable, Union, Any

from langchain.callbacks import get_openai_callback
from promptengineers.core.config.llm import ACCEPTED_OLLAMA_MODELS, ACCEPTED_OPENAI_MODELS, OpenAIModels
from promptengineers.llms.services.langchain.callbacks import AgentStreamCallbackHandler
from promptengineers.llms.services.langchain.chains import ChainService
from promptengineers.llms.strategies import OllamaStrategy, OpenAIStrategy, ModelContext
from promptengineers.llms.utils import retrieve_chat_messages, retrieve_system_message, get_chat_history
from promptengineers.models.message import SystemMessage, UserMessage, AssistantMessage
from promptengineers.retrieval.strategies import VectorstoreContext
from promptengineers.stream.utils import token_stream
from promptengineers.prompts.templates import get_retrieval_template

############################################################################
## Prompt Engineers AI - Retrieval HTTP Chat
############################################################################
async def langchain_http_retrieval_chat(
	messages: list[Union[SystemMessage, UserMessage, AssistantMessage]],
	model: str = OpenAIModels.GPT_3_5_TURBO_16K.value,
	temperature: float or int = 0.9,
	vectorstore: VectorstoreContext = None,
	openai_api_key: str = None,
) -> (str, Any):
	"""Send a message to the chatbot and yield the response."""
	filtered_messages = retrieve_chat_messages(messages)
	# Retrieve the chat history
	chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2])) ## TODO: Fix this
	# Retrieve the system message
	system_message = retrieve_system_message(messages)
	# Create the model
	if model in ACCEPTED_OPENAI_MODELS:
		model_service = ModelContext(strategy=OpenAIStrategy(api_key=openai_api_key))
	else:
		raise NotImplementedError(f"Model {model} not implemented")

	model = model_service.chat(
		model_name=model,
		temperature=temperature,
		streaming=False
	)
	with get_openai_callback() as cb:
		# Retrieve the conversation
		chain = ChainService(model).conversation_retrieval(
			system_message=system_message, vectorstore=vectorstore, chat_history=chat_history
		)
		# Begin a task that runs in the background.
		response = chain(filtered_messages[-1], return_only_outputs=True)
	return response, cb

############################################################################
## Prompt Engineers AI - Retrieval Stream Chat
############################################################################
async def langchain_stream_retrieval_chat(
	messages: list[Union[SystemMessage, UserMessage, AssistantMessage]],
	model: OpenAIModels,
	temperature: float = 0.9,
	vectorstore: VectorstoreContext = None,
	openai_api_key: str = None,
	ollama_base_url: str = None,
) -> AsyncIterable[str]:
	"""Send a message to the chatbot and yield the response."""
	filtered_messages = retrieve_chat_messages(messages)
	# Retrieve the chat history
	chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
	# Retrieve the system message
	system_message = retrieve_system_message(messages)
	# Create the callback
	callback = AgentStreamCallbackHandler()
	# Create the model
	if model in ACCEPTED_OPENAI_MODELS:
		model_service = ModelContext(strategy=OpenAIStrategy(api_key=openai_api_key))
		llm = model_service.chat(
			model_name=model,
			temperature=temperature,
			streaming=True,
			callbacks=[callback]
		)
		# Retrieve the conversation
		qa_chain = ChainService(llm).conversation_retrieval(
			system_message=system_message,
			chat_history=chat_history,
			vectorstore=vectorstore,
			callbacks=[callback]
		)
		runnable = qa_chain.astream_log(filtered_messages[-1])
		docs_processed = False
		async for chunk in runnable:
			operation = chunk.ops[0]['value']
			# print(operation)
			async for chunk in runnable:
				operation = chunk.ops[0]['value']
				return_output = False

				if isinstance(operation, dict):
					docs = operation.get('documents')
					if docs:
						for doc in docs:
							yield token_stream({
								'page_content': doc.page_content,
								'metadata': doc.metadata,
							}, 'doc')
						docs_processed = True  # Set this flag when docs are processed

				# Check if docs have been processed at least once
				if docs_processed:
					return_output = True

				if operation and isinstance(operation, str) and return_output:
					yield token_stream(operation)
		yield token_stream()

	elif model in ACCEPTED_OLLAMA_MODELS:
		if ollama_base_url:
			strategy = OllamaStrategy(base_url=ollama_base_url)
		else:
			strategy = OllamaStrategy()
		model_service = ModelContext(strategy=strategy)
		callback = AgentStreamCallbackHandler()
		llm = model_service.chat(
			model_name=model,
			temperature=temperature,
			streaming=True,
			callbacks=[callback]
		)
		question = filtered_messages[-1]
		template = get_retrieval_template(system_message)
		history = get_chat_history(chat_history)
		documents = vectorstore.similarity_search(question)
		context = '\n'.join([doc.page_content for doc in documents])
		prompt = template.format(
			chat_history=history,
			context=context,
			question=question
		)
		# Yield the tokens as they come in.
		for token in llm._stream(prompt):
			yield token_stream(token.text)
		yield token_stream()
	else:
		raise NotImplementedError(f"Model {model} not implemented")


"""Agent Chat"""
import ujson
from typing import Any

from langchain.callbacks import get_openai_callback
from promptengineers.core.config.llm import ACCEPTED_OPENAI_MODELS
from promptengineers.llms.services.langchain.callbacks import AgentStreamCallbackHandler
from promptengineers.llms.services.langchain.chains import ChainService
from promptengineers.llms.strategies import OpenAIStrategy, ModelContext
from promptengineers.llms.utils import retrieve_chat_messages, retrieve_system_message
from promptengineers.stream.utils import token_stream

############################################################################
## Prompt Engineers AI - Agent HTTP Chat
############################################################################
async def langchain_http_agent_chat(
    messages,
    model: str,
    tools,
    temperature: float or int = 0.0,
    openai_api_key: str = None,
) -> (str, Any):
    """Send a message to the chatbot and yield the response."""
    filtered_messages = retrieve_chat_messages(messages)
    # Retrieve the chat history
    chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
    # Retrieve the system message
    system_message = retrieve_system_message(messages)
    # Create the model
    if model in ACCEPTED_OPENAI_MODELS:
        model_service = ModelContext(strategy=OpenAIStrategy(api_key=openai_api_key))
    else:
        raise NotImplementedError(f"Model {model} not implemented")

    model = model_service.chat(
        model_name=model,
        temperature=temperature,
        streaming=False
    )
    with get_openai_callback() as cb:
        # Retrieve the conversation
        chain = ChainService(model).agent_with_tools(system_message=system_message,
                                                    chat_history=chat_history,
                                                    tools=tools)
        # Begin a task that runs in the background.
        response = await chain.ainvoke(filtered_messages[-1])
    return response, cb


############################################################################
## Prompt Engineers AI - Agent Stream Chat
############################################################################
async def langchain_stream_agent_chat(
    messages,
    model: str,
    tools,
    temperature: float or int = 0.9,
    openai_api_key: str = None,
):
    """Send a message to the chatbot and yield the response."""
    filtered_messages = retrieve_chat_messages(messages)
    # Retrieve the chat history
    chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
    # Retrieve the system message
    system_message = retrieve_system_message(messages)
    # Create the model
    if model in ACCEPTED_OPENAI_MODELS:
        model_service = ModelContext(strategy=OpenAIStrategy(api_key=openai_api_key))
    else:
        raise NotImplementedError(f"Model {model} not implemented")

    callback = AgentStreamCallbackHandler()
    model = model_service.chat(
        model_name=model,
        temperature=temperature,
        streaming=True,
        callbacks=[callback]
    )
    query = {'input': filtered_messages[-1], 'chat_history': chat_history}
    agent_executor = ChainService(model).agent_with_tools(system_message=system_message,
                                                        chat_history=chat_history,
                                                        tools=tools,
                                                        callbacks=[callback])
    runnable = agent_executor.astream_log(query)
    async for chunk in runnable:
        operation = chunk.ops[0]['value']
        if operation:
            if type(operation) == str:
                filled_chunk = operation
                if filled_chunk:
                    yield token_stream(filled_chunk)
            else:
                generations = operation.get('generations', False)
                if generations:
                    function_call = generations[0][0].get('message', {}).additional_kwargs.get('function_call', {})
                    tool = function_call.get('name', None)
                    if tool:
                        yield token_stream(tool, 'tool')
                    args = function_call.get('arguments', None)
                    if args:
                        if type(args) == str:
                            tool_args = ujson.loads(args)
                        else:
                            tool_args =  ujson.loads(args)['__arg1']

                        yield token_stream(f"Invoking: `{tool}` with `{tool_args}`", 'log')
    yield token_stream()



# import pickle
from abc import ABC, abstractmethod
# from io import BytesIO

from promptengineers.retrieval.services import PineconeService, RedisService


# Define the strategy interface
class VectorStoreStrategy(ABC):
    @abstractmethod
    def add(self, loaders, chunk_size: int = 1000, chunk_overlap: int = 100):
        pass

    @abstractmethod
    def load(self):
        pass

#########################################################################
## FAISS Strategy
#########################################################################
# class FaissStrategy(VectorstoreStrategy):
#     def __init__(self, _s3_access_key, _s3_secret_key, bucket_name, path):
#         self.s3_access_key = _s3_access_key
#         self.s3_secret_key = _s3_secret_key
#         self.bucket_name = bucket_name
#         self.path = path

#     def load(self):
#         f = BucketService(
#             self.s3_access_key,
#             self.s3_secret_key
#         ).retrieve_file(
#             bucket=self.bucket_name,
#             path=self.path
#         )
#         with BytesIO(f.read()) as file:
#             vectorstore = pickle.load(file)
#         return vectorstore

#########################################################################
## Pinecone Strategy
#########################################################################
class PineconeStrategy(VectorStoreStrategy):
    def __init__(
        self,
        api_key: str,
        env: str,
        namespace: str,
        index_name: str,
        embeddings = None,
    ):
        self.api_key = api_key
        self.env = env
        self.namespace = namespace
        self.index_name = index_name
        self.embeddings = embeddings
        self.service = PineconeService(
			api_key=self.api_key,
			env=self.env,
			index_name=self.index_name,
		)

    def list(self):
        return self.service.describe_index_stats()

    def add(self, loaders, chunk_size=1000, chunk_overlap=100):
        return self.service.from_documents(
            loaders=loaders,
            embeddings=self.embeddings,
            namespace=self.namespace,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap
        )

    def load(self):
        return self.service.from_existing(
            embeddings=self.embeddings,
            namespace=self.namespace
        )

#########################################################################
## Pinecone Strategy
#########################################################################
class RedisStrategy(VectorStoreStrategy):
    def __init__(
        self,
        redis_url: str,
        index_name: str,
        index_schema: dict = {"page_content": "TEXT", "metadata": "HASH"},
        embeddings = None,
    ):
        self.redis_url = redis_url
        self.index_name = index_name
        self.index_schema = index_schema
        self.embeddings = embeddings,
        self.service = RedisService(
			redis_url=self.redis_url,
			index_name=self.index_name,
            embeddings=self.embeddings
		)

    def list(self):
        return self.service.list_indexes()

    def add(self, loaders, chunk_size=1000, chunk_overlap=100):
        return self.service.add_docs(
            loaders=loaders,
            chunk_size=chunk_size,
            chunk_overlap=chunk_overlap
        )

    def load(self):
        return self.service.from_existing(
            schema=self.index_schema
        )

#########################################################################
## Strategy Context
#########################################################################
class VectorstoreContext:
    def __init__(self, strategy: VectorStoreStrategy):
        self.strategy = strategy

    def set_strategy(self, strategy: VectorStoreStrategy):
        self.strategy = strategy

    def add(self, loaders, chunk_size: int = 1000, chunk_overlap: int = 100):
        return self.strategy.add(loaders, chunk_size, chunk_overlap)

    def load(self):
        return self.strategy.load()



"""Chat Controller"""
import asyncio
from typing import Any, AsyncIterable

# from fastapi import Request
from langchain.chains import LLMChain
from langchain.callbacks import AsyncIteratorCallbackHandler, get_openai_callback

from promptengineers.core.config.llm import ACCEPTED_OLLAMA_MODELS, ACCEPTED_OPENAI_MODELS
from promptengineers.llms.services.langchain.callbacks import AgentStreamCallbackHandler
from promptengineers.llms.strategies import OllamaStrategy, OpenAIStrategy, ModelContext
from promptengineers.llms.utils import retrieve_chat_messages, retrieve_system_message, get_chat_history
from promptengineers.stream.utils import token_stream, wrap_done
from promptengineers.prompts.templates import get_system_template

##############################################################################
## Normal Chat
##############################################################################
async def langchain_http_chat(
    messages: list[dict],
    model: str,
    temperature: float or int = 0.0,
    openai_api_key: str = None,
    ollama_base_url: str = None,
) -> (str, Any):
    # Retrieve the chat messages
    filtered_messages = retrieve_chat_messages(messages)
    # Retrieve the chat history
    chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
    # Retrieve the system message
    system_message = retrieve_system_message(messages)
    # Check allowed
    if model in ACCEPTED_OPENAI_MODELS:
        model_service = ModelContext(strategy=OpenAIStrategy(api_key=openai_api_key))
    # Construct the model
    model = model_service.chat(
        model_name=model,
        temperature=temperature,
        streaming=False
    )
    query = {'question': filtered_messages[-1], 'chat_history': chat_history}
    with get_openai_callback() as cb:
        # Retrieve the conversation
        chain = LLMChain(llm=model, prompt=get_system_template(system_message))
        # Begin a task that runs in the background.
        response = chain.run(query)
    return response, cb

##############################################################################
## Langchain Chat Stream
##############################################################################
async def langchain_stream_chat(
    messages: list[dict],
    model:str,
    temperature: float or int = 0.0,
    openai_api_key: str = None,
    ollama_base_url: str = None,
) -> AsyncIterable[str]:
    """Send a message to the chatbot and yield the response."""
    filtered_messages = retrieve_chat_messages(messages)
    # Retrieve the chat history
    chat_history = list(zip(filtered_messages[::2], filtered_messages[1::2]))
    # Retrieve the system message
    system_message = retrieve_system_message(messages)

    # Create the model
    if model in ACCEPTED_OPENAI_MODELS:
        # Get Tokens
        model_service = ModelContext(strategy=OpenAIStrategy(api_key=openai_api_key))
        callback = AsyncIteratorCallbackHandler()
        model = model_service.chat(
            model_name=model,
            temperature=temperature,
            streaming=True,
            callbacks=[callback]
        )
        query = {'question': filtered_messages[-1], 'chat_history': chat_history}
        # Retrieve the conversation
        chain = LLMChain(llm=model, prompt=get_system_template(system_message))
        # Begin a task that runs in the background.
        task = asyncio.create_task(wrap_done(
            chain.acall(query),
            callback.done),
        )
        # Yield the tokens as they come in.
        async for token in callback.aiter():
            yield token_stream(token)
        yield token_stream()
        await task
    elif model in ACCEPTED_OLLAMA_MODELS:
        # Get Tokens
        if ollama_base_url:
            strategy = OllamaStrategy(base_url=ollama_base_url)
        else:
            strategy = OllamaStrategy()
        model_service = ModelContext(strategy=strategy)
        callback = AgentStreamCallbackHandler()
        llm = model_service.chat(
            model_name=model,
            temperature=temperature,
            streaming=True,
            callbacks=[callback]
        )
        template = get_system_template(system_message)
        prompt = template.format(
            context=get_chat_history(chat_history),
            question=filtered_messages[-1]
        )
        # Yield the tokens as they come in.
        for token in llm._stream(prompt):
            yield token_stream(token.text)
        yield token_stream()


import os
from promptengineers.retrieval.factories import EmbeddingFactory, LoaderFactory, RetrievalFactory
from promptengineers.retrieval.strategies import VectorstoreContext
from promptengineers.retrieval.utils import get_links

OPENAI_API_KEY = os.environ.get('OPENAI_API_KEY')
PINECONE_KEY = os.environ.get('PINECONE_API_KEY')
PINECONE_ENV = os.environ.get('PINECONE_ENV')
PINECONE_INDEX = os.environ.get('PINECONE_INDEX')

INDEX_PROVIDER = 'pinecone'
INDEX_NAME = 'default'
NAMESPACE = '63f0962f9a09c84c98ab6caf::formio'
EMBEDDING_LLM = 'text-embedding-ada-002'
LOADERS = [
	{
		"type": "copy",
	  	"text": "In a quaint village, an elderly man discovered a mysterious, "
				"ancient book in his attic. Each night, he read a chapter, and the "
				"next day, miracles began happening in the village. Unbeknownst to "
				"the villagers, their fortunes were being written by the man's dreams, "
				"guided by the magical book."
	},
	{
		"type": "web_base",
	  	"urls": ['https://adaptive.biz']
	},
	{
		"type": "website",
	  	"urls": ['https://adaptive.biz'] # Retrieves subpages
	},
	{
		"type": "youtube",
	  	"urls": ['https://www.youtube.com/watch?v=GyllRd2E6fg']
	}
]

# Generate Embeddings
embedding = EmbeddingFactory(EMBEDDING_LLM, OPENAI_API_KEY)

# Choose the appropriate vector search provider strategy for Pinecone
retrieval_provider = RetrievalFactory(
	INDEX_PROVIDER, 
	embedding.create_embedding(), 
	{
		'api_key': PINECONE_KEY, 
		'env': PINECONE_ENV, 
		'index_name': INDEX_NAME, 
		'namespace': NAMESPACE
	}
)

# Create loaders and process each
all_loaders = []
for loader_config in LOADERS:
	try:
		loader = LoaderFactory.create(loader_type=loader_config['type'], loader_config=loader_config)
		all_loaders.append(loader)
	except ValueError as e:
		print(f"Error creating loader: {e}")
	except Exception as e:
		print(f"Unexpected error: {e}")

# Assuming PineconeService and EmbeddingFactory have internal error handling
try:
	# Create a vector store service context
	vectostore_service = VectorstoreContext(retrieval_provider.create_strategy())
	vectorstore = vectostore_service.add(all_loaders)
	print(vectorstore)
except Exception as e:
	print(f"Error processing documents with Pinecone: {e}")


########################################################################
## Prompt Engineers AI - Agent Chat equipped w/ tools
########################################################################
import os
import asyncio

from promptengineers.chat import langchain_stream_agent_chat
from promptengineers.core.config.tools import AVAILABLE_TOOLS
from promptengineers.llms.utils import gather_tools
from promptengineers.retrieval.factories import EmbeddingFactory, RetrievalFactory
from promptengineers.retrieval.strategies import VectorstoreContext

# Environment Variables
OPENAI_API_KEY = os.environ.get('OPENAI_API_KEY')
PINECONE_KEY = os.environ.get('PINECONE_API_KEY')
PINECONE_ENV = os.environ.get('PINECONE_ENV')
PINECONE_INDEX = os.environ.get('PINECONE_INDEX')

# Retrieval Constants
INDEX_PROVIDER = 'pinecone'
INDEX_NAME = 'default'
NAMESPACE = '63f0962f9a09c84c98ab6caf::formio'
EMBEDDING_LLM = 'text-embedding-ada-002'

# Chat Constants
CHAT_LLM = 'gpt-4-1106-preview'
TEMPERATURE = 0.1
MESSAGES = [
    {
        'role': 'system', 
        'content': 'You are a powerful AI assistant, you are equipped with tools '
                    'to help you accomplish your tasks. Query the context when you need '
                    'additional information to complete your task. If the user query is not '
                    'related to the context then you can use the tools complete the task.'
    },
    {
        'role': 'user', 
        # 'content': 'What is 14125 compounded annually for 5 years at 4 percent for 23 years?' # Math Agent
        'content': 'Can you provide a react code sample to render a form in Form.io?'           # Retrieval Agent
    }
]

# Generate Embeddings
embedding = EmbeddingFactory(EMBEDDING_LLM, OPENAI_API_KEY)

# Choose the appropriate vector search provider strategy for Pinecone
retrieval_provider = RetrievalFactory(
    INDEX_PROVIDER, 
    embedding.create_embedding(), 
    {
        'api_key': PINECONE_KEY, 
        'env': PINECONE_ENV, 
        'index_name': INDEX_NAME, 
        'namespace': NAMESPACE
    }
)

# Create a vector store service context
vectostore_service = VectorstoreContext(retrieval_provider.create_strategy())

# Load the vectorstore using the service context
vectorstore = vectostore_service.load()

# Gather the tools
tools = gather_tools(tools=['math_tool'],
                    available_tools=AVAILABLE_TOOLS,
                    vectorstore=vectorstore,
                    plugins=[])

# Run the chat
async def main():
    response = langchain_stream_agent_chat(
        messages=MESSAGES,         
        model=CHAT_LLM,
        tools=tools,
        temperature=TEMPERATURE, 
        openai_api_key=OPENAI_API_KEY
    )
    async for data in response:
        print(data)

asyncio.run(main())


########################################################################
## Prompt Engineers AI - Retrieval Augmented Generation (RAG)
########################################################################
import os
import asyncio

from promptengineers.chat import langchain_stream_retrieval_chat
from promptengineers.retrieval.factories import EmbeddingFactory, RetrievalFactory
from promptengineers.retrieval.strategies import VectorstoreContext

# Environment Variables
OPENAI_API_KEY = os.environ.get('OPENAI_API_KEY')
PINECONE_KEY = os.environ.get('PINECONE_API_KEY')
PINECONE_ENV = os.environ.get('PINECONE_ENV')
PINECONE_INDEX = os.environ.get('PINECONE_INDEX')

# Retrieval Constants
INDEX_PROVIDER = 'pinecone'
INDEX_NAME = 'default'
NAMESPACE = '63f0962f9a09c84c98ab6caf::formio'
EMBEDDING_LLM = 'text-embedding-ada-002'

# Chat Constants
CHAT_LLM = 'gpt-3.5-turbo'
TEMPERATURE = 0.9
MESSAGES = [
    {
        'role': 'system', 
        'content': 'You are a helpful document retrieval AI, '
                    'use the context to answer the user queries.'
    },
    {
        'role': 'user', 
        'content': 'Can you summarize the context?'
    }
]

# Generate Embeddings
embedding = EmbeddingFactory(EMBEDDING_LLM, OPENAI_API_KEY)

# Choose the appropriate vector search provider strategy for Pinecone
retrieval_provider = RetrievalFactory(
    INDEX_PROVIDER, 
    embedding.create_embedding(), 
    {
        'api_key': PINECONE_KEY, 
        'env': PINECONE_ENV, 
        'index_name': INDEX_NAME, 
        'namespace': NAMESPACE
    }
)

# Create a vector store service context
vectostore_service = VectorstoreContext(retrieval_provider.create_strategy())

# Load the vectorstore using the service context
vectorstore = vectostore_service.load()

# Run the chat
async def main():
    response = langchain_stream_retrieval_chat(
        messages=MESSAGES,         
        model=CHAT_LLM, 
        temperature=TEMPERATURE, 
        vectorstore=vectorstore,
        openai_api_key=OPENAI_API_KEY
    )
    async for data in response:
        print(data)

asyncio.run(main())


from pydantic import BaseModel, Field

class SystemMessage(BaseModel):
	"""A message from the system."""
	role: str = Field(...)
	content: str = Field(...)

	__config__ = {
		"json_schema_extra": {
			"example": {
				"role": "system",
				"content": "You are a helpful assistant."
			}
		}
	}

class UserMessage(BaseModel):
	"""A message from the user."""
	role: str = Field(...)
	content: str = Field(...)

	__config__ = {
		"json_schema_extra": {
			"example": {
				"role": "user",
				"content": "Who won the 2001 world series?"
			}
		}
	}

class AssistantMessage(BaseModel):
	"""A message from the assistant."""
	role: str = Field(...)
	content: str = Field(...)

	__config__ = {
		"json_schema_extra": {
			"example": {
				"role": "assistant",
				"content": "The arizona diamondbacks won the 2001 world series."
			}
		}
	}


from promptengineers.models.settings import ChatSettings, CHAT_SETTINGS

class ReqBodySettings(ChatSettings): # pylint: disable=too-few-public-methods

	__config__ = {
		"json_schema_extra": {
			"example": CHAT_SETTINGS
		}
	}


import nest_asyncio
from langchain.document_loaders import (CSVLoader, DirectoryLoader, GitbookLoader,
										PyPDFLoader, TextLoader, JSONLoader, RecursiveUrlLoader,
										ReadTheDocsLoader, DataFrameLoader, UnstructuredHTMLLoader, 
										UnstructuredMarkdownLoader,
										UnstructuredURLLoader, WebBaseLoader,
										YoutubeLoader, SitemapLoader, BlockchainDocumentLoader)
from langchain.document_loaders.blockchain import BlockchainType
from promptengineers.core.config import ALCHEMY_API_KEY
from promptengineers.retrieval.loaders import CopyPasteLoader

nest_asyncio.apply()

class LoaderFactory:
    LOADER_CLASSES = {
        'gitbook': GitbookLoader,
        'web_base': WebBaseLoader,
        'website': RecursiveUrlLoader,               # Alias for 'web_base
        'youtube': YoutubeLoader,
        'polygon': BlockchainDocumentLoader,
        'ethereum': BlockchainDocumentLoader,
        'sitemap': SitemapLoader,
        'urls': UnstructuredURLLoader,          # Requires `unstructred` pip package (2.07 GB)
        'copy': CopyPasteLoader,
        'txt': TextLoader,
        'html': UnstructuredHTMLLoader,			# Requires `unstructred` pip package (2.07 GB)
        'md': UnstructuredMarkdownLoader,       # Requires `unstructred` pip package (2.07 GB)
        'directory': DirectoryLoader,           # Requires `unstructred` pip package (2.07 GB)
        'csv': CSVLoader,
        'pdf': PyPDFLoader,
        'json': JSONLoader,
        'pandas': DataFrameLoader,              # Requires `pandas`
        'readthedocs': ReadTheDocsLoader,       # Requires `beautifulsoup4`
    }

    @staticmethod
    def create(
        loader_type: ('gitbook', 'web_base', 'website', 'youtube', 'polygon', 'ethereum', 'sitemap', 'urls', 'copy', 'txt', 'html', 'md', 'directory', 'csv', 'pdf', 'json', 'pandas', 'readthedocs'), 
        loader_config
    ):
        loader_class = LoaderFactory.LOADER_CLASSES.get(loader_type)
        if not loader_class:
            raise ValueError(f'Unsupported document loader type: {loader_type}')

        # Special handling for blockchain loaders
        if loader_type in {'polygon', 'ethereum'}:
            blockchain_type = BlockchainType.POLYGON_MAINNET if loader_type == 'polygon' else None
            return loader_class(
                contract_address=loader_config.get('contract_address', ''),
                blockchainType=blockchain_type,
                api_key=ALCHEMY_API_KEY
            )

        # Special handling for the 'copy' loader
        if loader_type == 'copy':
            return loader_class(text=loader_config.get('text'))

        if loader_type == 'pandas':
            return loader_class(loader_config.get('df'), 
                                page_content_column=loader_config.get('page_content_column'))
        
        if loader_type == 'readthedocs':
            return loader_class(path=loader_config.get('path'), features='html.parser')
        
        if loader_type == 'gitbook':
            urls = loader_config.get('urls', [])
            return loader_class(urls=urls[0], load_all_paths=True)

        # Handling for loaders that require URLs or file paths
        if loader_type in {'web_base', 'sitemap', 'website', 'urls'}:
            urls = loader_config.get('urls', [])
            return loader_class(urls[0])

        if loader_type == 'youtube':
            urls = loader_config.get('urls', [])
            return loader_class.from_youtube_url(urls[0], add_video_info=False)
        
        if loader_type == 'json':
            return loader_class(file_path=loader_config.get('file_path'),
                                jq_schema=loader_config.get('jq_schema'),
                                text_content=loader_config.get('text_content'),
                                json_lines=loader_config.get('json_lines'))

        # Handling for file-based loaders
        return loader_class(loader_config.get('file_path'))
